import json
import collections
import argparse
import tempfile
import sys
import logging
import os
import shutil
import re
import enum
import string
from typing import Tuple, List, Mapping
from ruamel.yaml import YAML

from vnm.consts import PROG, NAME, VAR_PREFIX, DESC

log = logging.getLogger(__name__)
if sys.version_info < (3, 6):
    log.critical('`%s` MUST be run with Python 3.6 or higher. You are running v%s', PROG, '.'.join(sys.version))
    sys.exit(1)

import urllib.request
import urllib.parse
from typing import Optional, List
from subprocess import check_call, check_output, call, CalledProcessError, Popen, PIPE
from wheel.wheelfile import WheelFile

from vnm.repos import *
from vnm.constraint import Constraint
from vnm.package import Package
from vnm.venv import activate, deactivate, setEnv, defined, unsetEnv

VENVDIR = os.path.abspath('.venv')

yaml = YAML(typ='rt')

def require_pipjson():
    if not os.path.isfile('pip.json') and not os.path.isfile('vnm.yml'):
        log.critical('I can\'t find pip.json nor vnm.yml!')
        sys.exit(1)

def require_venv():
    if 'VIRTUAL_ENV' not in os.environ:
        log.critical('This command must be run from a venv virtual environment. Did you run `%s activate`? (VIRTUAL_ENV not set.)', PROG)
        sys.exit(1)

def getPython():
    pyexe = 'python.exe' if os.name == 'nt' else 'python'
    subdir = 'Scripts' if os.name == 'nt' else 'bin'
    return os.path.abspath(os.path.join(VENVDIR, subdir, pyexe))

def update_pip(args):
    xml = ''
    with urllib.request.urlopen('https://pypi.org/rss/project/pip/releases.xml') as u:
        #<title>20.2b1</title>
        xml = u.read().decode('utf-8')
    #print(xml)
    m = re.search(r'<title>([0-9\.]+)</title>', xml)
    assert m is not None
    upstream_version = m[1]

    o = check_output([getPython(), '-m', 'pip', '--version']).decode('utf-8')
    m = re.search(r'pip ([^ ]+) from (.*)', o)
    assert m is not None
    pip_version = m[1]

    print('Installed pip version: '+pip_version)
    print('Newest pip version...: '+upstream_version)
    if upstream_version!=pip_version:
        print('Updating pip...')
        check_call([getPython(), '-m', 'pip', 'install', '-U', 'pip'])

class EPipOperation(enum.IntEnum):
    INSTALL = enum.auto()
    REINSTALL = enum.auto()
    UPGRADE = enum.auto()

class ConfigState(object):
    PIPJSON_VERSION = 1
    VNMYML_VERSION = 2
    STATE_FILENAME = 'vnm.state'
    STATE_VERSION = '202006122014'
    def __init__(self, args: Optional[argparse.Namespace] = None):
        self.as_json: bool = False
        self.dev_mode: bool = False
        self.packages: Mapping[str, Package] = {}
        self.dev_packages: Mapping[str, Package] = {}

    def load(self) -> None:
        self.as_json = False
        data = {}
        if os.path.isfile('vnm.yml'):
            with open('vnm.yml', 'r') as f:
                docs = list(yaml.load_all(f))
                if isinstance(docs[0], int):
                    if docs[0] == 1:
                        docs[0]={'VERSION': 2}
                v = docs[0]['VERSION']
                if v == self.VNMYML_VERSION:
                    data = docs[1]
        elif os.path.isfile('pip.json'):
            log.warning('pip.json is deprecated.')
            self.as_json=True
            with open('pip.json', 'r') as f:
                data = json.load(f)

        if os.path.isfile('vnm.state'):
            with open('vnm.state', 'r') as f:
                docs = list(yaml.load_all(f))
                if docs[0] == self.STATE_VERSION:
                    self.dev_mode = docs[2].get('dev-mode', False)

        pkgdata = data.get('packages', {})
        for k, v in pkgdata.items():
            pkg = Package()
            pkg.id = k
            pkg.deserialize(v)
            self.packages[k] = pkg

        pkgdata = data.get('dev-packages', {})
        for k, v in pkgdata.items():
            pkg = Package()
            pkg.id = k
            pkg.deserialize(v)
            self.dev_packages[k] = pkg

    def get_packages(self) -> Mapping[str, Package]:
        o = self.packages.copy()
        if self.dev_mode:
            o.update(self.dev_packages)
        return o

    def get_all_packages(self) -> Mapping[str, Package]:
        return {**self.packages, **self.dev_packages}

    def _serialize_pkg_dict(self, d):
        o = collections.OrderedDict()
        for k in sorted(d.keys()):
            o[k]=d[k].serialize()
        return dict(o)
    def saveVNM(self, dry_run: bool=False) -> None:
        if self.as_json:
            data = {
                'version': self.PIPJSON_VERSION,
            }
            if len(self.packages) > 0:
                data['packages'] = self._serialize_pkg_dict(self.packages)
            if len(self.dev_packages) > 0:
                data['dev-packages'] = self._serialize_pkg_dict(self.dev_packages)
            with open('pip.new.json' if dry_run else 'pip.json', 'w') as f:
                json.dump(data, f, indent=2)
        else:
            data = {}
            if len(self.packages) > 0:
                data['packages'] = self._serialize_pkg_dict(self.packages)
            if len(self.dev_packages) > 0:
                data['dev-packages'] = self._serialize_pkg_dict(self.dev_packages)
            with open('vnm.new.yml' if dry_run else 'vnm.yml', 'w') as f:
                yaml.dump_all([{'VERSION' : self.VNMYML_VERSION}, data], f)

    def saveState(self, dry_run: bool=False) -> None:
        data={}
        if self.dev_mode:
            data['dev-mode']=True
        with open(self.STATE_FILENAME + ('.new' if dry_run else ''), 'w') as f:
            yaml.dump_all([self.STATE_VERSION, 'Do not manually edit this file, nor commit it to your repository.', data], f)

def install_venv(args: argparse.Namespace, cfg: ConfigState, operation=EPipOperation.INSTALL) -> List[Package]:
    pipargs = []
    verb = 'Installing'
    if operation == EPipOperation.REINSTALL:
        pipargs += ['-I']
        verb = 'Reinstalling'
    elif operation == EPipOperation.UPGRADE:
        pipargs += ['-U']
        verb = 'Upgrading'

    if len(cfg.packages) == 0 and len(cfg.dev_packages) == 0:
        return
    print('Writing temporary requirements file...')
    workdir = os.getcwd()
    tempfilename = ''
    all_packages: List[Package] = []
    try:
        if os.path.isfile('BROKEN-requirements.txt'):
            os.remove('BROKEN-requirements.txt')
        with tempfile.NamedTemporaryFile(delete=False, mode='w') as f:
            tempfilename = f.name
            for pid, pkg in cfg.get_packages().items():
                f.write(pkg.toRequirement()+'\n')
        print(f'{verb} packages from {tempfilename}...')
        check_call([getPython(), '-m', 'pip', 'install']+pipargs+['-r', tempfilename])
        '''
        with open(os.path.join(workdir, 'requirements.txt'), 'w') as f:
            for pid, pkg in cfg.packages.items():
                f.write(pkg.toRequirement()+'\n')
                all_packages += [pkg]
        with open(os.path.join(workdir, 'requirements-dev.txt'), 'w') as f:
            for pid, pkg in cfg.dev_packages.items():
                f.write(pkg.toRequirement()+'\n')
                all_packages += [pkg]
        '''
    except CalledProcessError as cpe:
        if os.path.isfile(tempfilename):
            shutil.copy(tempfilename, 'BROKEN-requirements.txt')
    finally:
        if os.path.isfile(tempfilename):
            os.remove(tempfilename)
    return all_packages

def update_requirements(cfg, args, all_packages):
    p = Popen([getPython(), '-m', 'pip', 'freeze'], stdout=PIPE, stderr=PIPE)
    stdout, stderr = p.communicate()
    suffix = '-dev' if cfg.dev_mode else ''
    if p.returncode == 0:
        all_package_names = []
        pkgs = {}
        for pkg in cfg.get_all_packages().values():
            all_package_names += [pkg.id.lower()]
            pkgs[pkg.id.lower()] = pkg
        for line in stdout.splitlines():
            #print(line)
            #oline = line
            line = line.strip()
            p = Package(line.decode('utf-8'))
            pkg = None
            if p.id.lower() in all_package_names:
                pkg = pkgs[p.id.lower()]
            #else:
            #    f.write(f'#{p.toRequirement()}\n')
            if pkg is not None:
                if len(pkg.constraints) == 0:
                    pkg.constraints = p.constraints
                if isinstance(pkg.repo, GitRepo) and isinstance(p.repo, GitRepo) and pkg.repo.refid != p.repo.refid:
                    pkg.repo = p.repo
            #print(line, pkg.id if pkg is not None else None, json.dumps(pkg.serialize()) if pkg is not None else None)

        all_release_pkgs = [k.lower() for k in cfg.packages.keys()]
        all_pkgs = all_release_pkgs+[k.lower() for k in cfg.dev_packages.keys()]
        with open(f'requirements.new.txt' if args.dry_run else f'requirements.txt', 'w') as f:
            for pkg in pkgs.values():
                if pkg.id.lower() in all_release_pkgs:
                    f.write(f'{pkg.toRequirement()}\n')
        if cfg.dev_mode:
            with open(f'requirements-dev.new.txt' if args.dry_run else f'requirements-dev.txt', 'w') as f:
                for pkg in pkgs.values():
                    if pkg.id.lower() in all_pkgs:
                        f.write(f'{pkg.toRequirement()}\n')
def main():

    argp = argparse.ArgumentParser(prog=PROG, description=DESC)
    argp.add_argument('--dry-run', action='store_true', default=False, help='Don\'t actually do anything.')
    subp = argp.add_subparsers()

    p_upgrade = subp.add_parser('upgrade', aliases=['u'], help='Upgrade all packages defined in vnm.yml to their latest available versions.')
    p_upgrade.set_defaults(cmd=_cmd_upgrade)

    p_upgrade_config = subp.add_parser('upgrade-config', help='Upgrade from pip.json to vnm.yml.')
    p_upgrade_config.set_defaults(cmd=_cmd_upgrade_config)

    p_install = subp.add_parser('install', aliases=['i'], help='Install all packages defined in vnm.yml')
    p_install.add_argument('--dev', '-d', action='store_true', default=False, help='Install in development mode. (Adds dev-packages.)')
    p_install.set_defaults(cmd=_cmd_install)

    p_init = subp.add_parser('init', help='Set up an initial .venv and vnm.yml.')
    p_init.add_argument('--clobber', action='store_true', default=False, help="Forcefully reinstall .venv and vnm.yml.")
    p_init.set_defaults(cmd=_cmd_init)

    p_add = subp.add_parser('add', aliases=['a'], help='Add a dependency to vnm.yml')
    p_add.add_argument('package', nargs='+', help="A package name or pip URI.")
    p_add.add_argument('--dev', '-d', action='store_true', default=False, help='All packages provided are development packages.')
    p_add.add_argument('--editable', '-e', nargs='+', type=str, help="A package name or pip URI (installed with -e).")
    p_add.add_argument('--force-reinstall', action='store_true', default=False, help="Forcefully reinstall all packages.")
    p_add.set_defaults(cmd=_cmd_add)

    p_remove = subp.add_parser('remove', aliases=['rm'], help='Remove package(s)')
    p_remove.add_argument('package', nargs='+', help="Package names.")
    p_remove.add_argument('--dev', '-d', action='store_true', default=False, help='All packages provided are development packages.')
    p_remove.set_defaults(cmd=_cmd_remove)

    p_activate = subp.add_parser('activate', help='Activate virtual environment.')
    p_activate.set_defaults(cmd=_cmd_activate)

    p_run = subp.add_parser('run', aliases=['r'], help='Run a python script in the virtual environment.')
    p_run.add_argument('filename', help='The python script or command to run')
    p_run.add_argument('args', nargs=argparse.REMAINDER, help='The python script to run')

    p_dumpenv = subp.add_parser('dump-env', help='Dump current environment variables.')
    p_dumpenv.add_argument('--format', choices=['text', 'json'], default='text', help='Format to output in')
    p_dumpenv.set_defaults(cmd=_cmd_dump_env)

    args = argp.parse_args()

    if getattr(args, 'cmd', None) is None:
        argp.print_usage()
    else:
        args.cmd(args)

def _sort_dict(d: dict) -> dict:
    nd = collections.OrderedDict()
    for k in sorted(d.keys()):
        nd[k]=d[k]
    return dict(nd)

def _cmd_add(args) -> None:
    require_venv()

    cfg = ConfigState()
    cfg.load()

    newpkgs = []
    all_packages = []
    for packagedef in args.package:
        pkg = Package(packagedef)
        log.info(f'Adding %s...', pkg)
        if pkg.id not in (None, '') and pkg.id in cfg.packages:
            log.warning('Package %r exists in %s.', pkg.id, cfg.get_filename())
        if args.dev:
            cfg.dev_packages[pkg.id] = pkg
        else:
            cfg.packages[pkg.id] = pkg
        if (cfg.dev_mode and args.dev) or (not cfg.dev_mode and not args.dev):
            newpkgs += pkg.toPipArgs()

    cfg.saveState(args.dry_run)
    cfg.saveVNM(args.dry_run)

    # Same as _cmd_install().
    activate(VENVDIR)
    update_pip(args)
    all_packages = install_venv(args, cfg)
    deactivate(VENVDIR)

    update_requirements(cfg, args, all_packages)

def _cmd_remove(args) -> None:
    require_venv()

    cfg = ConfigState()
    cfg.load()

    rmpkgs = []
    all_packages = []
    for packagedef in args.package:
        pkg = Package(packagedef)
        log.info(f'Removing %s...', pkg)
        if pkg.id not in (None, '') and pkg.id not in cfg.get_all_packages():
            log.warning('Package %r not specified in vnm.yml/pip.json.', pkg.id)
        if pkg.id in cfg.packages:
            del cfg.packages[pkg.id]
        if pkg.id in cfg.dev_packages:
            del cfg.dev_packages[pkg.id]
        rmpkgs += [pkg.id]

    cfg.saveState(args.dry_run)
    cfg.saveVNM(args.dry_run)

    # Same as _cmd_install().
    activate(VENVDIR)
    update_pip(args)
    call([getPython(), '-m', 'pip', 'uninstall', '-y']+rmpkgs)
    all_packages = install_venv(args, cfg)
    deactivate(VENVDIR)

    update_requirements(cfg, args, all_packages)

def _cmd_upgrade(args):
    require_venv()
    require_pipjson()

    cfg = ConfigState()
    cfg.load()

    activate(VENVDIR)
    update_pip(args)
    all_packages = install_venv(args, cfg, operation=EPipOperation.UPGRADE)
    deactivate(VENVDIR)

    update_requirements(cfg, args, all_packages)

def _cmd_upgrade_config(args):
    require_pipjson()

    cfg = ConfigState()
    cfg.load()

    log.warning('Upgrading pip.json to vnm.yml...')
    if os.path.isfile('pip.json'):
        if not os.path.isfile('pip.json.bak'):
            log.info('Backing up pip.json to pip.json.bak...')
            shutil.copy('pip.json', 'pip.json.bak')
        else:
            log.info("pip.json.bak exists, skipping backup.")
        log.info('Removing pip.json...')
        os.remove('pip.json')
        log.info('Writing vnm.yml...')

    cfg.as_json = False
    cfg.saveState(args.dry_run)
    cfg.saveVNM(args.dry_run)

def _cmd_install(args):
    if not os.path.isfile('vnm.yml'):
        if not os.path.isfile('pip.json'):
            log.critical('vnm.yml is missing.  This project is not set up for %s.', NAME)
            return
        log.warning('pip.json is deprecated.')
    if not os.path.isfile(getPython()):
        log.warning('Virtual environment not installed. Initializing...')
        clear_cmd = [sys.executable, '-m', 'venv', '--clear', '.venv']
        log.info('Running %s...', ' '.join(clear_cmd))
        check_call(clear_cmd)
    #require_venv()
    require_pipjson()

    cfg = ConfigState()
    cfg.load()
    cfg.dev_mode = args.dev
    cfg.saveState()

    activate(VENVDIR)
    update_pip(args)
    all_packages = install_venv(args, cfg)
    deactivate(VENVDIR)

    update_requirements(cfg, args, all_packages)

def _cmd_init(args):
    if (os.path.isfile('pip.json') or os.path.isfile('vnm.yml')) and not args.clobber:
        log.critical('pip.json or vnm.yml exists, aborting.  If you REALLY want to re-init the project, please set --clobber.')
        return

    log.info('Cleaning up prior install...')
    if os.path.isfile('pip.json'):
        log.info('rm pip.json')
        os.remove('pip.json')
    if os.path.isfile('vnm.yml'):
        log.info('rm vnm.yml')
        os.remove('vnm.yml')
    clear_cmd = [sys.executable, '-m', 'venv', '--clear', '.venv']
    log.info('Running %s...', ' '.join(clear_cmd))
    check_call(clear_cmd)

    log.info('Writing vnm.yml...')
    cfg = ConfigState()
    cfg.saveVNM()
    cfg.saveState()

    #update_pip(args)
    print( '********************************************************************')
    print(f'.venv for {NAME} installed!')
    print( 'Next steps:')
    print(f'  1. Run `{PROG} activate` to activate the virtual environment.')
    print(f'  2. Add packages with `{PROG} add <package>` OR edit .')
    print( '  3. Add vnm.yml and requirements.txt to your VCS (if applicable).')
    print( '********************************************************************')

def _cmd_dump_env(args):
    data = {
        'os.environ':os.environ,
        'sys.path':sys.path,
    }
    if args.format == 'json':
        print(json.dumps(data))
    else:
        print('os.environ:')
        for key, value in os.environ.items():
            print(f'  {key}: {value!r}')
        print('sys.path:')
        for value in sys.path:
            print(f'  - {value}')

def _cmd_activate(args):
    if not os.path.isfile(getPython()):
        log.critical('Virtual environment not installed. Did you run `%s init`?', PROG)
        return
    if f'_{VAR_PREFIX}_ACTIVE' in os.environ:
        log.critical(f'You are already in a {NAME} virtual environment, you doofus.')
        sys.exit(1)
        return

    if os.name == 'nt':
        _cmd_activate_nt(args)
    else:
        _cmd_activate_bash(args)

def _cmd_activate_nt(args):
    activate(VENVDIR)
    oldprompt = os.environ['PROMPT']
    os.environ['PROMPT'] = f'[{NAME}] {oldprompt}'
    print('='*60)
    print('Great, all done.')
    print('You are now in a child cmd.exe shell with the virtual environment set up for you.')
    print('You can deactivate the environment at any time by using the `exit` command.')
    print('='*60)
    def hide_script(a):
        if os.path.isfile(a):
            if os.path.isfile(a+'bak'):
                os.remove(a+'.bak')
            shutil.move(a, a+'.bak')
    def unhide_script(a):
        if os.path.isfile(a):
            os.remove(a)
        shutil.move(a+'.bak', a)
    hide_script(os.path.join(VENVDIR, 'Scripts', 'deactivate.bat'))
    code = -1
    try:
        code = call(['cmd.exe'], shell=True)
    finally:
        print(f'{PROG}: cmd.exe exited with code {code}')
        unhide_script(os.path.join(VENVDIR, 'Scripts', 'deactivate.bat'))
        deactivate(VENVDIR)
        os.environ['PROMPT'] = oldprompt

def _cmd_activate_bash(args):
    filename = ''
    with tempfile.NamedTemporaryFile(mode='w', delete=False) as f:
        filename = f.name
        log.debug('OPENED %s for writing')

        def writeNDebug(line):
            log.debug('Writing %r...', line)
            f.write(line)

        writeNDebug('echo "Activating..."\n')

        writeNDebug(f'. ~/.bashrc\n')
        writeNDebug(f'. {VENVDIR}/bin/activate\n')

        # Don't want users escaping by the normal means
        writeNDebug(f'unset -f deactivate\n')
        writeNDebug('alias deactivate=exit\n')
        writeNDebug(f'export _{VAR_PREFIX}_ACTIVE=1\n')
        writeNDebug(f'export _{VAR_PREFIX}_PROCESS={os.getpid()}\n')
        writeNDebug(f'PS1="[{NAME}] ${{PS1}}"\n')

        bar = '='*60
        writeNDebug(f'echo "{bar}"\n')
        writeNDebug('echo "Great, all done."\n')
        writeNDebug('echo "You are now in a child bash shell with the virtual environment set up for you."\n')
        writeNDebug('echo "You can deactivate the environment at any time by using the \\`exit\\` command."\n')
        writeNDebug(f'echo "{bar}"\n')
    code = call(['/bin/bash', '--init-file', filename])
    print(f'{PROG}: bash exited with code {code}')
    os.remove(filename)


#def _cmd_deactivate(args):
#    from vnm.venv import deactivate
#    deactivate('.venv')

def _cmd_run(args):
    from vnm.venv import runFrom
    runFrom('.venv', args.filename, args.args)

if __name__ == '__main__':
    main()



NAME = 'VNM'
PROG = 'vnm'
VAR_PREFIX = 'VNM'
DESC = 'Virtual eNvironment Manager'
VERSION = [0, 1, 0, 'beta']
VERSION_STR = '.'.join([str(x) for x in VERSION[:2]])+(f'-{VERSION[3]}' if len(VERSION)>3 else '')



import re
import os
from typing import Optional, List
from vnm.repos import *
from vnm.constraint import Constraint
class Package(object):
    REGEX_CONSTRAINTS=re.compile(r'(?P<equality>(==|>=|<=|>|<))(?P<version>[0-9\.]+)')
    REGEX_PACKAGENAME=re.compile(r'^([a-zA-Z0-9_\-\.]+)')

    @classmethod
    def FromJSON(cls, data: dict) -> 'Package':
        p = cls()
        p.deserialize(data)
        return p

    def __init__(self, pkgID: Optional[str] = None):
        self.id: str = ''
        self.development_mode: bool = False # -e
        self.repo: Repo = None
        self.constraints: List[Constraint] = []
        if pkgID is not None:
            if pkgID.startswith('-e '):
                self.development_mode = True
                pkgID=pkgID[3:]

            if os.path.isfile(pkgID):
                self.repo = WheelRepo(pkgID)
                self.id = self.repo.getID()
            elif pkgID.startswith('git+'):
                self.repo = GitRepo(pkgID)
                self.id = self.repo.pkgid
            else:
                repo = None
                if ' @ ' in pkgID:
                    pkgID, repo = pkgID.split('@', 1)
                    pkgID = pkgID.strip()
                    repo = repo.strip()
                m = self.REGEX_PACKAGENAME.match(pkgID)
                assert m is not None, f'Invalid package name {pkgID!r}'
                self.id = m[1]
                if repo is not None:
                    assert repo.startswith('git+')
                    self.repo = GitRepo(repo, knownPackageID=self.id)

                for cm in self.REGEX_CONSTRAINTS.findall(pkgID):
                    c = Constraint()
                    c.setEquality(cm[1])
                    c.setVersion(cm[2])
                    self.constraints += [c]

    def __str__(self) -> str:
        o = f'{self.id}'
        o += f' ({self.repo})' if self.repo is not None else ''
        if self.development_mode:
            o += ' [developer mode (-e)]'
        return o

    def serialize(self) -> dict:
        o = {}
        if self.development_mode:
            o['development'] = True
        if self.repo is not None:
            o['repo'] = self.repo.serialize()
        if len(self.constraints) > 0:
            o['constraints'] = []
            for constraint in self.constraints:
                o['constraints'] += [constraint.serialize()]
        return o

    def deserialize(self, data: dict) -> None:
        self.development_mode = data.get('development', False)
        self.repo = None
        if 'repo' in data:
            #print(repr(data['repo']))
            rt = data['repo']['type']
            if rt == 'git':
                self.repo = GitRepo()
            elif rt == 'wheel':
                self.repo = WheelRepo()
            else:
                print(f"Error: Unknown repo.type={rt!r}")
            if self.repo is not None:
                self.repo.deserialize(data['repo'])
                #print(repr(self.repo), str(self.repo))
        if 'constraints' in data:
            for cdata in data['constraints']:
                c = Constraint()
                c.deserialize(cdata)
                self.constraints += [c]


    def toRequirement(self) -> str:
        o = ''
        if self.development_mode:
            o += '-e '
        if self.repo is not None:
            o += self.repo.toRequirement(self)
        else:
            o += self.id
        if self.repo is None or self.repo.shouldDisplayConstraints():
            o += ','.join([x.toRequirement() for x in self.constraints])
        return o

    def toPipArgs(self) -> List[str]:
        o = []
        if self.development_mode:
            o += ['-e']
        if self.repo is not None:
            o += self.repo.toPipArgs(self)
        if self.repo is None or self.repo.shouldDisplayConstraints():
            o += ','.join([x.toRequirement() for x in self.constraints])
        return o



import os
from subprocess import check_call
from vnm.consts import NAME, VAR_PREFIX
OLD_VENV_VAR = ''
def defined(varname):
    return varname in os.environ

def activate(venvdir) -> None:
    if os.name == 'nt':
        _activate_nt(venvdir)
    if os.name == 'posix':
        _activate_bash(venvdir)
    setEnv(f'_{VAR_PREFIX}_ACTIVE', '1')
    setEnv(f'_{VAR_PREFIX}_PROCESS', str(os.getpid()))

def deactivate(venvdir: str, nondestructive: bool = False) -> None:
    if os.name == 'nt':
        _deactivate_nt(venvdir, nondestructive)
    if os.name == 'posix':
        _deactivate_bash(venvdir, nondestructive)
    if defined(f'_{VAR_PREFIX}_ACTIVE'): unsetEnv(f'_{VAR_PREFIX}_ACTIVE')
    if defined(f'_{VAR_PREFIX}_PROCESS'): unsetEnv(f'_{VAR_PREFIX}_PROCESS')

def runFrom(venvdir: str, pythonfile: str, args: str):
    activate(venvdir)
    if os.path.isfile(pythonfile):
        pyexe = 'python.exe' if os.name == 'nt' else 'python'
        check_call([os.path.join(venvdir, 'bin', pyexe), pythonfile] + args)
    else:
        check_call(pythonfile + args)
    deactivate(venvdir)

def setEnv(varname: str, value: str) -> None:
    print(f'ENV: {varname} = {value}')
    os.environ[varname] = value

def unsetEnv(varname: str) -> None:
    print(f'ENV: {varname} (UNSET)')
    del os.environ[varname]

def getPrompt(venvdir) -> None:
    dirname = venvdir
    # Aspen shit
    if os.path.basename(venvdir) == '__':
        dirname = os.path.dirname(venvdir)
    dirname = os.path.basename(dirname)
    return f'({dirname}) '

def _activate_bash(venvdir: str) -> None:
    # unset irrelevant variables
    #deactivate nondestructive
    _deactivate_bash(venvdir, True)

    #VIRTUAL_ENV="__VENV_DIR__"
    #export VIRTUAL_ENV
    setEnv('VIRTUAL_ENV', venvdir)

    #_OLD_VIRTUAL_PATH="$PATH"
    setEnv('_OLD_VIRTUAL_PATH', os.environ['PATH'])
    #PATH="$VIRTUAL_ENV/__VENV_BIN_NAME__:$PATH"
    setEnv('PATH', os.environ['VIRTUAL_ENV']+'/bin:'+os.environ['PATH'])
    #export PATH

    # unset PYTHONHOME if set
    # this will fail if PYTHONHOME is set to the empty string (which is bad anyway)
    # could use `if (set -u; : $PYTHONHOME) ;` in bash
    #if [ -n "${PYTHONHOME:-}" ] ; then
    #    _OLD_VIRTUAL_PYTHONHOME="${PYTHONHOME:-}"
    #    unset PYTHONHOME
    #fi
    if defined('PYTHONHOME'):
        setEnv('_OLD_VIRTUAL_PYTHONHOME', os.environ['PYTHONHOME'])
        unsetEnv('PYTHONHOME')

    #if [ -z "${VIRTUAL_ENV_DISABLE_PROMPT:-}" ] ; then
    #    _OLD_VIRTUAL_PS1="${PS1:-}"
    #    if [ "x__VENV_PROMPT__" != x ] ; then
    #	PS1="__VENV_PROMPT__${PS1:-}"
    #    else
    #    if [ "`basename \"$VIRTUAL_ENV\"`" = "__" ] ; then
    #        # special case for Aspen magic directories
    #        # see http://www.zetadev.com/software/aspen/
    #        PS1="[`basename \`dirname \"$VIRTUAL_ENV\"\``] $PS1"
    #    else
    #        PS1="(`basename \"$VIRTUAL_ENV\"`)$PS1"
    #    fi
    #    fi
    #    export PS1
    #fi
    if not defined('VIRTUAL_ENV_DISABLE_PROMPT'):
        os.environ['_OLD_VIRTUAL_PS1'] = os.environ['PS1'] if defined('PS1') else '\\u@\\h \\w>'
        os.environ['PS1'] = getPrompt(venvdir)+os.environ.get('PS1', '\\u@\\h \\w>')

    # NOTE: This is already checked in _deactivate_posix(). Skipped
    # This should detect bash and zsh, which have a hash command that must
    # be called to get it to forget past commands.  Without forgetting
    # past commands the $PATH changes we made may not be respected
    #if [ -n "${BASH:-}" -o -n "${ZSH_VERSION:-}" ] ; then
    #    hash -r
    #fi

def _activate_nt(venvdir: str) -> None:
    setEnv('VIRTUAL_ENV', venvdir)

    if not defined('PROMPT'):
        setEnv('PROMPT', '$P$G')
    #if defined _OLD_VIRTUAL_PROMPT set PROMPT=%_OLD_VIRTUAL_PROMPT%
    if defined('_OLD_VIRTUAL_PROMPT'):
        setEnv('PROMPT', os.environ['_OLD_VIRTUAL_PROMPT'])
    #if defined _OLD_VIRTUAL_PYTHONHOME set PYTHONHOME=%_OLD_VIRTUAL_PYTHONHOME%
    if defined('_OLD_VIRTUAL_PYTHONHOME'):
        setEnv('PYTHONHOME', os.environ['_OLD_VIRTUAL_PYTHONHOME'])

    #set _OLD_VIRTUAL_PROMPT=%PROMPT%
    setEnv('_OLD_VIRTUAL_PROMPT', os.environ['PROMPT'])
    #set PROMPT=__VENV_PROMPT__%PROMPT%
    setEnv('PROMPT', getPrompt(venvdir) + os.environ['PROMPT'])
    #if defined PYTHONHOME set _OLD_VIRTUAL_PYTHONHOME=%PYTHONHOME%
    if defined('PYTHONHOME'):
        setEnv('_OLD_VIRTUAL_PYTHONHOME', os.environ['PYTHONHOME'])
        #set PYTHONHOME=
        unsetEnv('PYTHONHOME')

    #if defined _OLD_VIRTUAL_PATH set PATH=%_OLD_VIRTUAL_PATH%
    if defined('_OLD_VIRTUAL_PATH'):
        setEnv('PATH', os.environ['_OLD_VIRTUAL_PATH'])
    #if not defined _OLD_VIRTUAL_PATH set _OLD_VIRTUAL_PATH=%PATH%
    if not defined('_OLD_VIRTUAL_PATH'):
        setEnv('_OLD_VIRTUAL_PATH', os.environ['PATH'])

    #set PATH=%VIRTUAL_ENV%\__VENV_BIN_NAME__;%PATH%
    setEnv('PATH', os.environ['VIRTUAL_ENV']+'\\bin;'+os.environ['PATH'])

    #:END
    #if defined _OLD_CODEPAGE (
    #    "%SystemRoot%\System32\chcp.com" %_OLD_CODEPAGE% > nul
    #    set _OLD_CODEPAGE=
    #)

def _deactivate_nt(venvdir: str, nondestructive: bool = False) -> None:
    # venv currently doesn't have a deactivate script for NT, making it up
    # as we go.
    if defined('_OLD_VIRTUAL_PROMPT'):
        setEnv('PROMPT', os.environ['_OLD_VIRTUAL_PROMPT'])
        unsetEnv('_OLD_VIRTUAL_PROMPT')
    if not defined('PROMPT'):
        setEnv('PROMPT', '$P$G')
    if defined('_OLD_VIRTUAL_PYTHONHOME'):
        setEnv('PYTHONHOME', os.environ['_OLD_VIRTUAL_PYTHONHOME'])
        unsetEnv('_OLD_VIRTUAL_PYTHONHOME')
    if defined('_OLD_VIRTUAL_PATH'):
        setEnv('PATH', os.environ['_OLD_VIRTUAL_PATH'])
        unsetEnv('_OLD_VIRTUAL_PATH')
    if defined('VIRTUAL_ENV'):
        unsetEnv('VIRTUAL_ENV')

def _deactivate_bash(venvdir: str, nondestructive: bool = False) -> None:
    #deactivate () {
    # reset old environment variables
    #    if [ -n "${_OLD_VIRTUAL_PATH:-}" ] ; then
    #        PATH="${_OLD_VIRTUAL_PATH:-}"
    #        export PATH
    #        unset _OLD_VIRTUAL_PATH
    #    fi
    if defined('_OLD_VIRTUAL_PATH'):
        setEnv('PATH', os.environ['_OLD_VIRTUAL_PATH'])
        unsetEnv('_OLD_VIRTUAL_PATH')

    #    if [ -n "${_OLD_VIRTUAL_PYTHONHOME:-}" ] ; then
    #        PYTHONHOME="${_OLD_VIRTUAL_PYTHONHOME:-}"
    #        export PYTHONHOME
    #        unset _OLD_VIRTUAL_PYTHONHOME
    #    fi
    if defined('_OLD_VIRTUAL_PYTHONHOME'):
        setEnv('PYTHONHOME', os.environ['_OLD_VIRTUAL_PYTHONHOME'])
        unsetEnv('_OLD_VIRTUAL_PYTHONHOME')

    # This should detect bash and zsh, which have a hash command that must
    # be called to get it to forget past commands.  Without forgetting
    # past commands the $PATH changes we made may not be respected
    #    if [ -n "${BASH:-}" -o -n "${ZSH_VERSION:-}" ] ; then
    #        hash -r
    #    fi
    if defined('BASH') or defined('ZSH_VERSION'):
        # We can't run hash -r here, since hash won't affect the parent bash.
        print("IMPORTANT: You will need to run `hash -r` after this.")

    #    if [ -n "${_OLD_VIRTUAL_PS1:-}" ] ; then
    #        PS1="${_OLD_VIRTUAL_PS1:-}"
    #        export PS1
    #        unset _OLD_VIRTUAL_PS1
    #    fi
    if defined('_OLD_VIRTUAL_PS1'):
        os.environ['PS1'] = os.environ['_OLD_VIRTUAL_PS1']
        del os.environ['_OLD_VIRTUAL_PS1']

    #    unset VIRTUAL_ENV
    if defined('VIRTUAL_ENV'):
        unsetEnv('VIRTUAL_ENV')

    # We don't create this.
    #    if [ ! "${1:-}" = "nondestructive" ] ; then
    #        # Self destruct!
    #        unset -f deactivate
    #    fi

    #}



from typing import List
from wheel.wheelfile import WheelFile
import os
import shutil
import tempfile
import urllib.request
from vnm.repos.repo import Repo

class WheelRepo(Repo):
    TYPEID = 'wheel'

    def __init__(self, uri: str = ""):
        self.uri: str = uri

    def __str__(self):
        return f'Wheel: {self.uri}'

    def serialize(self) -> dict:
        data = super().serialize()
        data['uri'] = self.uri
        return data

    def deserialize(self, data: dict) -> None:
        super().deserialize(data)
        self.uri = data['uri']

    def toRequirement(self, parent: 'Package') -> str:
        return self.uri

    def toPipArgs(self, parent: 'Package') -> List[str]:
        return [self.uri]

    def shouldDisplayConstraints(self) -> bool:
        return False

    def getID(self) -> str:
        path = self.uri
        ntf = None
        try:
            if not os.path.isfile(self.uri):
                ntf = tempfile.NamedTemporaryFile(delete=False)
                path = ntf.name
                with urllib.request.urlopen(self.uri) as response:
                    shutil.copyfileobj(response, ntf)
            with WheelFile(path) as wf:
                return wf.parsed_filename.group('namever')
        finally:
            if ntf is not None:
                ntf.close()
                if os.path.isfile(path):
                    os.remove(path)



from typing import List
class Repo(object):
    TYPEID = ''
    def __init__(self):
        pass

    def serialize(self) -> dict:
        return {'type': self.TYPEID}

    def deserialize(self, data: dict) -> None:
        return

    def toRequirement(self, parent: 'Package') -> str:
        return ''

    def toPipArgs(self, parent: 'Package') -> List[str]:
        return []

    def shouldDisplayConstraints(self) -> bool:
        return True



import enum
import re
import sys
import urllib.parse
import string
from typing import List, Optional
from vnm.repos.repo import Repo
class EGitRefType(enum.IntEnum):
    NONE   = enum.auto()
    BRANCH = enum.auto()
    TAG    = enum.auto()
    COMMIT = enum.auto()
    BLOB   = enum.auto()

class GitRepo(Repo):
    TYPEID = 'git'
    REGEX_FORMAT_1 = re.compile(r'^git\+(?P<uri>[^@#]+)(@(?P<refid>[a-z0-9/\-_]+))?')
    REGEX_FORMAT_2 = re.compile(r'^git\+(?P<uri>[^@#]+)(@(?P<refid>[a-z0-9/\-_]+))?#egg=(?P<pkgid>[a-z0-9_\-]+)')
    REGEX_REFID_COMMON_TAGS = re.compile(r'^[vV]?(\d+\.)+\d+(\-(rc|alpha|beta)?\d*)$')
    LONG_COMMIT_LEN = 40 # 58186079583c71760fd6b1950bee5fa92e3cd364
    SHORT_COMMIT_LEN = 7 # I think
    def __init__(self, uri: str = "", knownPackageID: Optional[str] = None):
        self.uri: str = ''
        self.pkgid: str = knownPackageID
        self.refid: Optional[str] = None
        self.reftype: EGitRefType = EGitRefType.NONE

        super().__init__()

        def _err_bad_uri_format():
            print(f'Improper git repo definition string: {uri}')
            print(f'Proper format: git+schema://uri/to/repo.git[@branch-or-commit]#egg=packageid')
            print(f'               git+username@server:/path/to/repo.git[@branch-or-commit]#egg=packageid')
            sys.exit(1)

        if uri != '':
            if '#egg=' in uri:
                m = self.REGEX_FORMAT_2.match(uri)
                if m is None:
                    _err_bad_uri_format()
                else:
                    self.uri = m.group('uri')
                    if self.pkgid is None:
                        self.pkgid = m.group('pkgid')
                    self.refid = m.group('refid')
            else:
                m = self.REGEX_FORMAT_1.match(uri)
                if m is None:
                    _err_bad_uri_format()
                else:
                    self.pkgid = None
                    self.uri = m.group('uri')
                    self.refid = m.group('refid')
                    #print(self.refid)
                    uriinfo = urllib.parse.urlparse(self.uri)
                    if self.pkgid is None:
                        self.pkgid = uriinfo.path.split('/')[-1].split('.')[0]
            if self.refid is not None:
                # If the refid is all hex and the length of commits sha1s...
                if len(self.refid) == self.LONG_COMMIT_LEN and all(x in string.hexdigits for x in self.refid):
                    self.reftype = EGitRefType.COMMIT
                elif self.REGEX_REFID_COMMON_TAGS.match(self.refid) is not None:
                    self.reftype = EGitRefType.TAG
                else:
                    self.reftype = EGitRefType.BRANCH


    def __str__(self):
        o = f'Git: {self.uri}'
        if self.reftype != EGitRefType.NONE:
            o += ' @ '
            if self.reftype == EGitRefType.BLOB:
                o += f'blob {self.refid}'
            if self.reftype == EGitRefType.BRANCH:
                o += f'branch {self.refid!r}'
            if self.reftype == EGitRefType.TAG:
                o += f'tag {self.refid!r}'
            if self.reftype == EGitRefType.COMMIT:
                o += f'commit {self.refid}'
        return o

    def serialize(self) -> dict:
        data = super().serialize()
        data['uri'] = self.uri
        if self.reftype != EGitRefType.NONE:
            data[self.reftype.name.lower()] = self.refid
        return data

    def deserialize(self, data: dict) -> None:
        super().deserialize(data)
        self.uri = data['uri']
        if 'branch' in data:
            self.reftype = EGitRefType.BRANCH
            self.refid = data['branch']
        elif 'tag' in data:
            self.reftype = EGitRefType.TAG
            self.refid = data['tag']
        elif 'commit' in data:
            self.reftype = EGitRefType.COMMIT
            self.refid = data['commit']

    def toRequirement(self, parent: 'Package') -> str:
        line = 'git+'+self.uri
        if self.refid is not None:
            line += f'@{self.refid}'
        line += f'#egg={parent.id}'
        return line

    def toPipArgs(self, parent: 'Package') -> List[str]:
        return [self.toRequirement(parent)]

    def shouldDisplayConstraints(self) -> bool:
        return False



import enum
from typing import Optional
class EEqualityType(enum.IntEnum):
    EQUAL                 = enum.auto()
    GREATER_THAN_OR_EQUAL = enum.auto()
    LESS_THAN_OR_EQUAL    = enum.auto()
    LESS_THAN             = enum.auto()
    GREATER_THAN          = enum.auto()

    @staticmethod
    def ToRequirement(val):
        if val == EEqualityType.EQUAL:
            return '=='
        if val == EEqualityType.GREATER_THAN_OR_EQUAL:
            return '>='
        if val == EEqualityType.LESS_THAN_OR_EQUAL:
            return '<='
        if val == EEqualityType.GREATER_THAN:
            return '>'
        if val == EEqualityType.LESS_THAN:
            return '<'
        return '??'

class Constraint(object):
    def __init__(self, txt: Optional[str] = None):
        self.version: str = ''
        self.equality: EEqualityType = EEqualityType.EQUAL

    def setVersion(self, v: str) -> None:
        self.version = v

    def setEquality(self, eq: str) -> None:
        if eq in ('==', 'eq'):
            self.equality = EEqualityType.EQUAL
        elif eq in ('>=', 'gte'):
            self.equality = EEqualityType.GREATER_THAN_OR_EQUAL
        elif eq in ('<=', 'lte'):
            self.equality = EEqualityType.LESS_THAN_OR_EQUAL
        elif eq in ('>', 'gt'):
            self.equality = EEqualityType.GREATER_THAN
        elif eq in ('<', 'lt'):
            self.equality = EEqualityType.LESS_THAN

    def toRequirement(self) -> str:
        o = ''
        o += EEqualityType.ToRequirement(self.equality)
        o += str(self.version)
        return o

    def serialize(self) -> dict:
        return {
            'comparison': EEqualityType.ToRequirement(self.equality),
            'version': self.version,
        }

    def deserialize(self, data: dict) -> None:
        if isinstance(data, list):
            self.setEquality(data[0])
            self.setVersion(data[1])
        else:
            self.setVersion(data['version'])
            self.setEquality(data['comparison'])



from setuptools import setup, find_packages

README = ''
with open('README.md', 'r') as f:
    README = f.read()
setup(
    name='vnm',
    version='0.2.0',
    packages=find_packages(),
    license='MIT',
    author='Rob "N3X15" Nelson',
    author_email='nexisentertainment@gmail.com',
    long_description=README,
    long_description_content_type='text/markdown',
    entry_points={
        'console_scripts': [
            'vnm = vnm.__main__:main'
        ]
    },
    install_requires=['wheel', 'ruamel.yaml']
)



import platform
import json
import os
from subprocess import PIPE, DEVNULL, Popen
def source(filename):
    if platform.system() == 'Linux':
        source_bash(filename)

def source_bash(filename):
    proc = Popen(f'source {filename};vnm dump-env --json', executable='/bin/bash', shell=True)
    stdout = proc.communicate()
    data = json.loads(stdout)
    os.environ.clear()
    os.environ.update(data)
